//  MUColor.cpp version 1.5
//  yudit package - Unicode Editor for the X Window System (and Linux) 
//
//  Author: gsinai@iname.com (Gaspar Sinai)
//  GNU Copyright (C) 1997,1998,1999  Gaspar Sinai
// 
//  yudit version 1.5  Copyright(C) 30 November, 1999, Tokyo Japan  Gaspar Sinai
//  yudit version 1.4  Copyright(C) 25 November, 1999, Tokyo Japan  Gaspar Sinai
//  yudit version 1.3  Copyright(C)  5 April,    1999, Tokyo Japan  Gaspar Sinai
//  yudit version 1.2  Copyright(C) 10 December, 1998, Tokyo Japan  Gaspar Sinai
//  yudit version 1.1  Copyright(C) 23 August,   1998, Tokyo Japan  Gaspar Sinai
//  yudit version 1.0  Copyright(C) 17 May,      1998, Tokyo Japan  Gaspar Sinai
//  yudit version 0.99 Copyright(C)  4 April,    1998, Tokyo Japan  Gaspar Sinai
//  yudit version 0.97 Copyright(C)  4 February, 1998, Tokyo Japan  Gaspar Sinai
//  yudit version 0.95 Copyright(C) 10 January,  1998, Tokyo Japan  Gaspar Sinai
//  yudit version 0.94 Copyright(C) 17 December, 1997, Tokyo Japan  Gaspar Sinai
//  yudit version 0.9 Copyright (C)  8 December, 1997, Tokyo Japan  Gaspar Sinai
//  yutex version 0.8 Copyright (C)  5 November, 1997, Tokyo Japan  Gaspar Sinai
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

/*
 * Need to have access to the guts
 */
#include <X11/IntrinsicP.h>
#include <X11/CoreP.h>
#include <X11/StringDefs.h>
#include <iostream.h>
#include "MUColor.h"

static Colormap		privateC=0;
static Bool		isPrivate;
static Colormap		original=0;

static Boolean cvtStringToPixel (Display *dpy, XrmValuePtr args, Cardinal *nargs, 
	XrmValuePtr from, XrmValuePtr to, XtPointer junk);

static int CAInit = 0;

/*
 * If the desired color is farer then allocate color if there are more
 * than COLOR_FREE free cells are available.
 */
#define COLOR_DISTANCE 10
#define COLOR_FREE     10

static Status CXAllocColor (Display *display, Colormap colormap, XColor *screen_in_out);
static Status initColors (Display *display, Colormap colormap);
static Status myXAllocColor (Display *display, Colormap colormap, XColor *screen_in_out);
static int getFreeEntries ();

MUColor::MUColor(Widget widgetIn)
{
	widget = widgetIn;
}

MUColor::~MUColor()
{
}

void
MUColor::setNamedColor(const char* name)
{
	Colormap	myMap;
	Visual*		visual;
	XColor		scol;
	XColor		ecol;

	visual = DefaultVisualOfScreen (XtScreen (widget));
	myMap = DefaultColormapOfScreen (XtScreen (widget));

	XtVaGetValues (widget, XtNcolormap, &myMap, 0);
	if (visual->c_class !=  PseudoColor)
	{
		if (XAllocNamedColor(XtDisplay (widget), myMap, (String) name, &scol, &ecol) == 0)
		{
			cerr << "warn: could not allocate color " 
				<< name << ".\n";
		}
		else
		{
			colorName = name;
			pixelVle =  scol.pixel;
		}
		return;
	}
	if (XAllocNamedColor(XtDisplay (widget), myMap, (String) name, &scol, &ecol) == 0)
	{
		if (XParseColor(XtDisplay (widget), myMap, (String) name, &scol) == 0)
		{
			cerr << "warn: could not allocate color " 
				<< name << ".\n";
		}
		else if (CXAllocColor(XtDisplay (widget), myMap, &scol) == 0)
		{
			cerr << "warn: could not allocate color " 
				<< name << ".\n";
		}
		else
		{
			colorName = name;
			pixelVle =  scol.pixel;
		}
		return;
	}
	colorName = name;
	pixelVle =  scol.pixel;
}

/*
 * These flags indicate the availability of a
 * particular color cell.
 */
typedef enum
{
        FLAG_FREE = 0,
        FLAG_SHARED,
        FLAG_ISHARE,
        FLAG_IALLOC,
        FLAG_USED
} FLAG_TYPE;

typedef struct 
{
        Display         *display;
        Screen          *screen;
	Colormap        defaultColormap;
	Visual          *visual;
        int             cells;
        XColor          *colors;
        char            *flags;
	Pixel		*pixels;
	int		free;
	int		distance;
	int		allocLimit;
} colorMap;

static colorMap map;

void
MYuditSetColorMap (Widget w, Bool install)
{
	static XtConvertArgRec  colorArgs[] =
	{
		{ XtWidgetBaseOffset,
		(XtPointer) XtOffsetOf (WidgetRec, core.colormap),
			sizeof(Colormap) },
	};

	map.display = XtDisplay (w);;
	map.screen = XtScreen (w);
	map.visual = DefaultVisualOfScreen (map.screen);
	if (map.visual->c_class !=  PseudoColor) return;

	XtVaGetValues (w, XtNcolormap, &original, 0);

	isPrivate = install;

	/*
	 * Currently I implemented only string conversion
	 */
	XtSetTypeConverter (XtRString, XtRPixel,
		(XtTypeConverter) cvtStringToPixel,
		colorArgs, XtNumber (colorArgs),
		XtCacheRefCount, 0);

	if (!install) return;
	privateC = XCopyColormapAndFree (XtDisplay (w), original);
	XtVaSetValues (w, XtNcolormap, privateC, 0);
	return;
}

static Boolean
cvtStringToPixel (Display *dpy, XrmValuePtr args, Cardinal *nargs, XrmValuePtr from, 
	XrmValuePtr to, XtPointer junk)
{
	String		name = (String)(from->addr);
	XColor		scol, ecol;
	static Pixel	op;
	Colormap	cmap;

	cmap = *((Colormap *)args[0].addr);

	if (XAllocNamedColor(dpy, cmap, name, &scol, &ecol) == 0)
	{
		if (XParseColor(dpy, cmap, name, &scol) == 0)
			return False;
		if (CXAllocColor(dpy, cmap, &scol) == 0)
			return False;
	}
	op = scol.pixel;
	if (to->addr == 0)
	{
		to->addr = (char*) &op;
	}
	else
	{
		if (to->size < sizeof(Pixel))
			return False;
		memcpy(to->addr, &op, sizeof(Pixel));
	}
	to->size = sizeof(Pixel);
	return True;
}

static Status
CXAllocColor (Display *display, Colormap colormap, XColor *screen_in_out)
{
	/*
	 * If we are using private colormap we should not try out complicated
	 * scheme.
 	 */
	if (isPrivate)
	{
		return (XAllocColor(display, colormap, screen_in_out));
	}
	if (CAInit == 0)
	{
		if (initColors (display, colormap) == False)
		{
			return False;
		}
	}
	if (display != map.display) 
	{
		return XAllocColor (display, colormap, screen_in_out);
	}
	return myXAllocColor (display, colormap, screen_in_out);
}

static Status
initColors (Display *display, Colormap colormap)
{
	int	i;

	map.pixels = 0;
	map.colors = 0;
	map.flags = 0;

	/* display screen and visual are initialozed already */

	CAInit = 1;

	if (map.visual->c_class !=  PseudoColor)
	{
		map.display = 0;
		return True;
	}

	map.defaultColormap = colormap;
	map.cells = CellsOfScreen (map.screen);
	map.colors = new XColor [map.cells];
	map.flags = new char [map.cells];
	map.pixels = new Pixel [map.cells];

	if (map.colors == 0 || map.flags == 0 || map.pixels ==0)
	{
		map.display = 0;
		return False;
	}

	for (i=0; i<map.cells; i++)
	{
		map.colors[i].pixel = i;
		map.flags[i] = FLAG_FREE;
	}
	map.distance = COLOR_DISTANCE;
	map.allocLimit = COLOR_FREE;
	getFreeEntries ();
	return True;
}

static Status
myXAllocColor (Display *display, Colormap colormap, XColor *screen_in_out)
{
	int     bestMatch;
	long    distance, d, distance2;
	long    r,g,b;
	XColor  color;
	int     i;
	int     fallbacks = 0;

	fallbacks = map.cells;

Fallback:
	if (fallbacks != map.cells)
	{
		getFreeEntries ();
	}
	if (fallbacks-- == 0)
	{
		return False;
	}
	color = *screen_in_out;
	bestMatch=-1;
	distance=0;
	for (i=0;i<map.cells;i++)
	{
		if (map.flags[i] == FLAG_SHARED 
			|| map.flags[i] == FLAG_ISHARE
			|| map.flags[i] == FLAG_IALLOC)
		{
			/* divide to prevent overflow error */
			r = (map.colors[i].red>>2)-(color.red>>2);
			g = (map.colors[i].green>>2)-(color.green>>2);
			b = (map.colors[i].blue>>2)-(color.blue>>2);

			d = r*r+g*g+b*b;

			if (bestMatch<0 || d<distance)
			{
				distance = d;
				bestMatch = i;
			}
		}
	}

	/* distance is measured from 0-1000 */
	distance2 = distance/805;
	if (bestMatch > 0 && distance2 <= (map.distance) * (map.distance))
	{
		color = map.colors[bestMatch];
		/*
		 * Possibly we already allocated this color.
		 * If not - it is only bad luck -racing with another app
		 */ 
		if (!XAllocColor (map.display, map.defaultColormap, &color))
		{
			/* Try it again */
			map.flags[bestMatch] = FLAG_USED;
			goto Fallback;
		}
		/*
		 * Oops. We happened to tumble into some allocated cells.
		 * Mark them as used.
		 */ 
		if (color.pixel != map.colors[bestMatch].pixel)
		{
			map.flags[bestMatch] = FLAG_USED;
			/* 
			 * We can not make this - other apps may not use
			 * our library
			 */
			/*
			 * XFreeColors(map.display,map.defaultColormap,
			 *	&color.pixel,1,0L);
			 */
			goto Fallback;
		}
		*screen_in_out = color;
		map.flags[bestMatch] = FLAG_ISHARE;
		if (distance == 0) return True;
		return True;
	}

	/* Allocate map.colors only if the free entires are few */
	if (map.free > 0 && map.allocLimit <= map.free)
	{
		/*
	 	* Try to allocate it
	 	*/
		color = *screen_in_out;
		color.flags = DoRed | DoGreen | DoBlue;
		if (!XAllocColor (map.display, map.defaultColormap, &color))
		{
			/* Try it again */
			goto Fallback;
		}
		*screen_in_out = color;
		map.flags[color.pixel] = FLAG_IALLOC;
		getFreeEntries ();
		return  True;
	}

	if (bestMatch < 0)
	{
		cerr << "Can not find any shreable map.colors.\n";
		return False;
	}
	color = map.colors[bestMatch];

	/*
	 * Possibly we already allocated this color.
	 * If not - it is only bad luck -racing with another app
	 */ 
	if (!XAllocColor (map.display, map.defaultColormap, &color))
	{
		/* Try it again */
		map.flags[bestMatch] = FLAG_USED;
		goto Fallback;
	}
	/*
	 * Oops. We happened to tumble into some allocated cells.
	 * Mark them as used.
	 */ 
	if (color.pixel != map.colors[bestMatch].pixel)
	{
		map.flags[bestMatch] = FLAG_USED;
		/* 
		 * We can not make this - other apps may not use
		 * our library
		 */
		/*
		 * XFreeColors(map.display,map.defaultColormap,
		 *	&color.pixel,1,0L);
		 */
		goto Fallback;
	}
	map.flags[color.pixel] = FLAG_ISHARE;
	*screen_in_out = color;
	return True;
}

static int
getFreeEntries ()
{
	int	i;

	XQueryColors (map.display,map.defaultColormap,map.colors,map.cells);

	for (i=0; i<map.cells; i++)
	{
		map.colors[i].flags = DoRed | DoGreen | DoBlue;
		if (map.flags[i] != FLAG_USED
			&& map.flags[i] != FLAG_ISHARE
			&& map.flags[i] != FLAG_IALLOC)
		{
			map.flags[i] = FLAG_SHARED;
		}
	}
	i=0;
	while( XAllocColorCells(map.display, map.defaultColormap,
		False, 0L, 0, &map.pixels[i], 1 ))
	{
		if (map.flags[map.pixels[i]] != FLAG_ISHARE
			&& map.flags[map.pixels[i]] != FLAG_IALLOC)
		{
			map.flags[map.pixels[i]] = FLAG_FREE;
		}
		i++;
	}
	map.free = i;
	XFreeColors(map.display, map.defaultColormap, map.pixels,i,0L);
	XSync( map.display, 0 );
	return (map.free);
}

